/* REP_CLIENT2.C

   29-JAN-1999, David Mathog, Biology Division, Caltech

   REP_CLIENT2 is a program which can be used to POST a file to a web site,
   reformatting the human readable file into the format the web server expects.
   It then sends the web server's response to stdout.

   REP_CLIENT2.C is a modified version of the NCBI's REP_CLIENT.C program.
   Not sure who the author was, but the URL where I got it was:

      ftp://ftp.ncbi.nlm.nih.gov/blast/blasturl/urldemo.tar

   This variant preprocesses the input so that
   <space> -> <+>
   <EOL>   -> %OD
   etc.
   and all other appropriate character translations are made.  This has to
   be done or POST won't work for most sites.

   The program reads from stdin a file in this format:


POST /cgi-bin/target HTTP/1.0
www.target.machine.com
FIELD1=1
FIELD2=
FIELD3=some text string
FIELD4=
TEXTFIELD_BEGIN
a string
another string
the last string in the text area
TEXTFIELD_END
FIELD5=IS_SET

   The special fields TEXTFIELD_BEGIN, TEXTFIELD_END set things up so that
   TEXTAREA type data is formatted correctly.  Note that CHECKFIELD should
   only be present in the input file if SET.

Build it with 

   Modified

   $ cc/standard=ansi89/prefix=all/define=_POSIX_C_SOURCE rep_client2.c
   $ link rep_client2.obj

   use it like:

   $ rrr:==$wherever:rep_client2
   $ define/user sys$input p.in
   $ define/user sys$output out.txt
   $ rrr

  Porting method (from Unix)

  1.  replaced all includes with those from FETCH_HTTP.C 
  2.  added ctype.h for isprint

  Total port time - 5 minutes!

  3.  modification.  Pass in 

        POST_COMMAND "POST /cgi-bin/BLAST/nph-blast_report HTTP/1.0"
        HOST         "www.ncbi.nlm.nih.gov"

   In the command file       

  4. (for rep_client2)  Added a translation table, textarea support.

*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#if defined(VMS) || defined(__VMS)
#include <unixio.h>
#include <unistd.h>
#include <types.h>
#include <socket.h>
#include <in.h>
#include <lib$routines.h>
#else
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif
#include <errno.h>
#include <netdb.h>

/* add a prototype */

int SendBlastRequest(char *);
char * http_sanitize(char * query_string); /* new for rep_client2.c */

#define HUGE_STRING_LEN 1024


/*#define ENGINE_HOST  "blast.planet.nlm.nih.gov"*/
/* #define PROGRAM_NAME "nph-blast_report"
#define ENGINE_HOST  "www.ncbi.nlm.nih.gov"
#define ENGINE_PORT  80
*/

int  Port=80;
char *blast_engine_host; 
char *post_command;

#define TRUE  1
#define FALSE 0

static int connect_handler(int fd);
static void error_handler(int error);

char buffer[HUGE_STRING_LEN];

char *ReadFileInMemory(FILE *fd, int len, int filter);

main(int argc, char **argv)
{
    char *query_string;
    char *ptr;
    char *sptr;

    query_string = ReadFileInMemory(stdin, 0, FALSE);

    
    if(query_string == NULL) {
        printf("No data read. No requests send\n");
        exit(1);
    }

    /* set up the key web parameters */

    blast_engine_host=query_string;
    for (ptr=query_string; *ptr!='\n'; ptr++){}
    *ptr='\0';
    ptr++;
    post_command=ptr;
    for ( ; *ptr!='\n'; ptr++){}
    ptr--;
    *ptr='\n';
    ptr++;
    *ptr='\0';
    ptr++;

    sptr = http_sanitize(ptr); /* new for rep_client2.c */

    if(!SendBlastRequest(sptr)) {
        printf("Error sending request to BLAST server\n");
        exit(1);
    }
    exit(0);
}

int SendBlastRequest(char *query_string)
{
    int  Length;
    int sock;
    char myhostname[128];
    char *fragout;
    static char *remainder;
    
    struct sockaddr_in server;
    struct hostent *hp;
    int  nbytes, i;
    fd_set fds;
    
    if(query_string == NULL || (Length = strlen(query_string)) == 0)
        return FALSE;
    
    /* Creating new socket */
    
    memset(&server, '\0', sizeof(server));
    
    if((sock = socket(AF_INET, SOCK_STREAM, 0)) <= 0) {
        error_handler(-2);
        exit(1);
    }
    
    /* Getting address of remote host from DNS */
    
    if((hp = gethostbyname(blast_engine_host)) == NULL) {
        error_handler(-3);
        printf("\nError in detection of host address\n");
        exit(1);
    }
    
    memcpy((void *) &server.sin_addr, hp->h_addr, hp->h_length);
    
    server.sin_family = hp->h_addrtype;    /* Family is AF_INET */
    server.sin_port = htons(Port);         /* Server port to connect to */

    /* Connecting to the WWW server */  
    
    if(connect(sock, (struct sockaddr *) &server, sizeof(server)) < 0) {
        error_handler(-4);
        printf("Errno is %d\n", errno);
        return FALSE;
    }

    /* Sockets are connected. Now sending request */

    write(sock, post_command, strlen(post_command));

    gethostname(myhostname, sizeof(myhostname));

    sprintf(buffer, "User-Agent: SpecialClient from %s\n", myhostname);
    write(sock, buffer, strlen(buffer));
    sprintf(buffer, "Connection: Keep-Alive\n");
    write(sock, buffer, strlen(buffer));
    sprintf(buffer, "Content-type: application/x-www-form-urlencoded\n");
    write(sock, buffer, strlen(buffer));
    sprintf(buffer, "Content-Length: %d\n\n", Length);
    write(sock, buffer, strlen(buffer));
    write(sock, query_string, Length);
    
    /* Now waiting for data back and printing data on STDOUT */

    while (TRUE) {
        
        FD_ZERO(&fds);
        FD_SET(sock, &fds);
        
        if ((i = select(sock+1, &fds, NULL, NULL, NULL)) != 0) {
            if (FD_ISSET(sock, &fds)) { /* Socket activity */
                
                /* We read from socket and write to stdout */
                
                if((nbytes = read(sock, buffer, HUGE_STRING_LEN)) > 0) {

                /* the block of text which came back may or may
                   not end with a new line, that is, a text line
                   may have been split.  If one was split, save it
                   as remainder. */

                    buffer[nbytes]='\0';
                    for( fragout=&buffer[0], remainder=fragout ;
                         *fragout != '\0'   ;
                         fragout++ )
                    {
                      if(*fragout=='\n'){
                         *fragout='\0';
                         fprintf(stdout,"%s\n",remainder);
                         remainder=fragout;
                         remainder++;
                      }
                    }
                    if(fragout != remainder){ /* some left over to span to next string */
                      fprintf(stdout,"%s",remainder);
                    }
/*
This method won't work on OpenVMS - it assumes that the output file is a
"pile of bytes", which you cannot guarantee on OpenVMS. So if a line is
split into 2 different transfer blocks, write will cause it to be split
on output into 2 lines.  The method above looks for these cases and
is careful to use fprintf with no newline to handle them.

                       if(write(1, buffer, nbytes) == EOF)  
                        break;
*/
                } else if(nbytes == 0) {
                    break;
                } else {
                    break;
                }                
            } else {
                error_handler(-5);
                break; /* This must not happend ... */
            }
        } else {
            error_handler(-6);
            break;  /* Timeout expired if any ... */
        }
    }
    
    close(sock);
    fflush(stdout);
    close(0);
    return TRUE;
}
char *ReadFileInMemory(FILE *fd, int len, int filter)
{
    int     bytes = 0;
    char    *in_buff;
    int     new_size = HUGE_STRING_LEN;
    int     buff_len = 0;
    register int i;
  
    if(fd == NULL)
        return NULL;

    if(len == 0) {  
        /* initial allocation of memory */
    
        if((in_buff = malloc(HUGE_STRING_LEN)) == NULL) {
            printf("Error in allocating memory\n");
            return NULL;
        }
    
        while ((bytes = fread(in_buff + buff_len, 1,
                              HUGE_STRING_LEN, fd)) > 0) {
            new_size += bytes;
            buff_len += bytes;
            
            if ((in_buff = realloc(in_buff, new_size)) == NULL) {
                printf("Error in reallocating memory\n");
                return NULL;
            }
        }
        in_buff[buff_len] = '\0';
        
    } else {
    
        in_buff = malloc(len+1);
    
        while(buff_len < len) {
            if((bytes = fread(in_buff+buff_len, 1, len, fd)) == 0
               || bytes == len) {
                buff_len += bytes;
                break; /* EOF or done*/
            } else if (bytes < 0) {
                /* some error setting may be here */
                return NULL;
            }
            buff_len += bytes;
        }
        in_buff[buff_len] = '\0';
    }
  
    /* some filtering of non-printing characters */
    
    if(filter) {
        for(i = 0; i < buff_len; i++) {
            if (!isprint(in_buff[i]) && !isspace(in_buff[i]))
                in_buff[i] = ' ';
        }
    }
  
    return(in_buff);
}

static void error_handler(int error)
{
    printf("Error code is %d<BR>");
    fflush(stdout);
    return;
}

/* new for rep_client2.c */

/* is_exact string looks for an exact match, including
case, of a string pattern in string (offset by OFFSET)
and the TEST pattern, up to the length of the TEST pattern.
(String may be longer than TEST, but not vice versa).  In this application
the offset is always 1 because there is an \n before the pattern we are
looking for */ 

int is_exact_string(char *test, char *string, int offset){
char *p1;
char *p2;
int result;

  result = 0;
  for (p1=string; offset > 0 ; p1++, offset--){}
  for (p2=test ; ;p1++,p2++){
     if(*p1 != *p2){ /* strings don't match exactly */
       if(*p2 == '\0'){  /* but matched until end of TEST, so ok */
          result = 1;
       }
       break; 
     }
     if(*p2 == '\0'){      /* strings match, and are same length TEST */
       result = 1;
       break;
     }
  }
  return result;
}

char * http_sanitize(char * iffy_string){
char * to;
char * hold;
char * from;
char * tfrom;
char  pass[]="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890_=-.*";
int last;
enum EOL_STATES {EOL_NORMAL,EOL_EAT_FIRST,EOL_CRLF};
enum EOL_STATES eol_status;
   
   to = malloc(strlen(iffy_string)*3*sizeof(char));  /* big enough to hold any translation*/
   if(to == NULL){
        error_handler(-7);
        printf("\nCould not allocate enough memory to process POST request\n");
        exit(1);
   }
   hold = to;
   last = 0;
   *to = '\0';
   eol_status=EOL_NORMAL;
   /* copy over the first two lines verbatim */
   for (from=iffy_string ; *from != '\0' ; from++){
      last = 0;
      if(strchr(pass,*from) == NULL){ /* not in pass list */
        switch (*from) {
          case ' ':
            *to = '+';
            to++;
            break;
          case '\n':
            if(is_exact_string("TEXTFIELD_BEGIN",from,1)){
               for(from++; *from != '\n';from++){}    /* skip \nTEXTFIELD_BEGIN */
               eol_status=EOL_EAT_FIRST;
            }
            if(is_exact_string("TEXTFIELD_END"  ,from,1)){
               for(from++; *from != '\n';from++){}    /* skip \nTEXTFIELD_END */
               eol_status=EOL_NORMAL;
            }
            switch (eol_status) {
               case EOL_NORMAL:
                 last = 1;
                 *to = '&';
                 to++;
                 break;
               case EOL_EAT_FIRST:  /* merge field name with following data */
                 eol_status=EOL_CRLF;
                 break;
               case EOL_CRLF:
                 strcat(to,"%0D%0A");
                 to++;to++;to++;to++;to++;to++;
                 break;
            }
            break;
          default:  /* translate it to 2 digit hex (%xx) */
            sprintf(to,"%%%02.2x",*from);
            to++;
            to++;
            to++;
            break;
        }
      }
      else {  /* pass *from -> *to */
        *to = *from;
        to++;
      }
   }
   if(last == 1){
     to--;
     *to='\n';
     to++;
   }
   *to='\0';   /* it might already be, but be doubly sure */
   return hold;
}
